@@ -7,12 +7,14 @@ import logging
 import sys
 import threading
 from collections import OrderedDict
+import socket
 
-from .ddp import get_status
+from .ddp import get_status, search
 from .oauth import prompt as oauth_prompt
 from .register import register
-from .session import Session, SessionAsync
+from .session import Session
 from .util import add_profile, add_regist_data, get_profiles, write_profiles
+from .feedback import Controller
 
 NEW_PROFILE = "New Profile"
 CANCEL = "Cancel"
@@ -24,27 +26,75 @@ _LOGGER = logging.getLogger(__name__)
 
 def main():
     """Main entrypoint."""
-    parser = argparse.ArgumentParser(description='Start Remote Play.')
-    parser.add_argument('host', type=str, help="IP address of Remote Play host")
-    parser.add_argument('-r', '--resolution', default="720p", type=str, choices=RESOLUTIONS, help="Resolution to use")
-    parser.add_argument('-f', '--fps', default="high", type=str, choices=FPS_CHOICES, help="Max FPS to use")
-    parser.add_argument('--register', action="store_true", help='Register with Remote Play host.')
+    if "-l" in sys.argv or "--list" in sys.argv:
+        show_devices()
+        return
+    parser = argparse.ArgumentParser(description="Start Remote Play.")
+    parser.add_argument(
+        "host", type=str, default="", help="IP address of Remote Play host"
+    )
+    # parser.add_argument(
+    #     "-r",
+    #     "--resolution",
+    #     default="720p",
+    #     type=str,
+    #     choices=RESOLUTIONS,
+    #     help="Resolution to use",
+    # )
+    # parser.add_argument(
+    #     "-f",
+    #     "--fps",
+    #     default="high",
+    #     type=str,
+    #     choices=FPS_CHOICES,
+    #     help="Max FPS to use",
+    # )
+
+    # Just to show in help
+    parser.add_argument(
+        "-l",
+        "--list",
+        action="store_true",
+        help="List Devices",
+    )
+    parser.add_argument(
+        "--register", action="store_true", help="Register with Remote Play host."
+    )
 
     args = parser.parse_args()
     host = args.host
-
     should_register = args.register
-    resolution = args.resolution
-    fps = args.fps
-    if fps.isnumeric():
-        fps = int(fps)
+    resolution = "360p"
+    fps = 30
+    # resolution = args.resolution
+    # fps = args.fps
+    # if fps.isnumeric():
+    #     fps = int(fps)
 
+    try:
+        socket.gethostbyname(host)
+    except socket.gaierror:
+        print(f"\nError: Could not find host with address: {host}\n\n")
+        parser.print_help()
+        return
     if should_register:
         register_profile(host)
         return
     cli(host, resolution, fps)
 
 
+def show_devices():
+    """Print all devices."""
+    devices = search()
+    print(f"\nFound {len(devices)} devices:\n")
+    for device in devices:
+        print(
+            f"\tIP Address: {device.get('host-ip')}\n"
+            f"\tName: {device.get('host-name')}\n"
+            f"\tType: {device.get('host-type')}\n\n"
+        )
+
+
 def select_profile(profiles: dict, use_single: bool, get_new: bool) -> str:
     """Return profile name."""
     name = NEW_PROFILE
@@ -79,9 +129,8 @@ def register_profile(host: str):
     if not status:
         print("Host is not reachable")
         return
-    if status.get("status_code") != 200:
+    if status.get("status-code") != 200:
         return
-    mac_address = status.get("host-id")
 
     profiles = get_profiles()
     if profiles:
@@ -122,40 +171,41 @@ def register_profile(host: str):
 
 
 def cli(host: str, resolution: str, fps: str):
+    """Start CLI."""
     profiles = get_profiles()
     if profiles:
         name = select_profile(profiles, True, False)
         profile = profiles[name]
-        session = SessionAsync(host, profile, resolution=resolution, fps=fps)
+        session = Session(host, profile, resolution=resolution, fps=fps)
         async_start(session, cb_curses)
     else:
         _LOGGER.info("No Profiles")
 
 
 def cb_curses(session, status: bool):
+    """Start Curses."""
     if status:
         instance = CLIInstance(session)
-        worker = threading.Thread(
-            target=curses.wrapper,
-            args=(start, instance)
-        )
+        worker = threading.Thread(target=curses.wrapper, args=(start, instance))
         worker.start()
     else:
         _LOGGER.error("Session Failed to Start: %s", session.error)
 
 
-def async_start(session, cb: callable):
+def async_start(session, callback: callable):
+    """Sync method for starting session."""
     loop = asyncio.get_event_loop()
     session.loop = loop
-    task = loop.create_task(async_start_session(session, cb))
+    task = loop.create_task(async_start_session(session, callback))
     loop.run_until_complete(task)
     loop.run_forever()
 
 
-async def async_start_session(session, cb: callable):
+async def async_start_session(session, callback: callable):
+    """Start Session."""
     status = await session.start()
     if status:
-        cb(session, status)
+        callback(session, status)
     else:
         asyncio.get_event_loop().stop()
 
@@ -165,7 +215,7 @@ def start(stdscr, instance):
     instance.run(stdscr)
 
 
-class CLIInstance():
+class CLIInstance:
     """Emulated Keyboard."""
 
     MAP = {
@@ -193,6 +243,7 @@ class CLIInstance():
 
     def __init__(self, session: Session):
         self._session = session
+        self.controller = Controller()
         self.stdscr = None
         self.last_key = None
         self.map = self.MAP
@@ -207,24 +258,24 @@ class CLIInstance():
         curses.init_pair(5, curses.COLOR_GREEN, curses.COLOR_BLACK)
 
     def _show_mapping(self):
-        self.stdscr.addstr('\n')
+        self.stdscr.addstr("\n")
         _key_mapping = OrderedDict()
-        _key_mapping.update({'Key': 'Action'})
+        _key_mapping.update({"Key": "Action"})
         _key_mapping.update(self.map)
         item = 0
         for key, value in _key_mapping.items():
-            if key == '\n':
-                key = 'KEY_ENTER'
+            if key == "\n":
+                key = "KEY_ENTER"
             self.stdscr.addstr(key, curses.color_pair(5))
-            self.stdscr.addstr(' : ')
+            self.stdscr.addstr(" : ")
             self.stdscr.addstr(value, curses.color_pair(4))
             item += 1
             if item >= 4:
                 item = 0
-                self.stdscr.addstr('\n')
+                self.stdscr.addstr("\n")
             else:
-                self.stdscr.addstr(' | ')
-        self.stdscr.addstr('\n\n')
+                self.stdscr.addstr(" | ")
+        self.stdscr.addstr("\n\n")
         self._pos = self.stdscr.getyx()
         self.stdscr.clrtobot()
         self.stdscr.refresh()
@@ -235,26 +286,35 @@ class CLIInstance():
         self.stdscr.clear()
         self.stdscr.refresh()
         self.stdscr.addstr(
-            0, 0,
+            0,
+            0,
             "Remote Play - Interactive mode, press 'q' to exit\n",
         )
         self._show_mapping()
 
     def run(self, stdscr):
         """Run CLI Instance."""
+        self.controller.connect(self._session)
+        self.controller.start()
         self.stdscr = stdscr
         self._init_color()
         self.stdscr.scrollok(True)
         self._init_window()
         timeout = 50
         self.stdscr.timeout(timeout)
+        self.stdscr.clrtobot()
+        _last = ""
         while not self._session.state == Session.STATE_STOP:
-            self.stdscr.refresh()
+            self._init_window()
+            if _last:
+                self._write_str(_last, 5)
             try:
-                self._handle_key(self.stdscr.getkey())
+                key = self._handle_key(self.stdscr.getkey())
+                if key:
+                    _last = key
             except curses.error:
                 if self.last_key is not None:
-                    self._session.controller.button(self.last_key, "release")
+                    self.controller.button(self.last_key, "release")
                     self.last_key = None
 
     def _write_str(self, text, color=1):
@@ -266,14 +326,18 @@ class CLIInstance():
     def _handle_key(self, key):
         key = self.map.get(key)
         if key and self.last_key is None:
-            self._write_str(key, 5)
             self.last_key = key
-            self._session.controller.button(key, "press")
             if key == "QUIT":
+                self._write_str(key, 3)
+                self.stdscr.refresh()
                 self._session.stop()
-                self._session.loop.stop()
+                self._session.loop.call_soon_threadsafe(self._session.loop.stop)
                 sys.exit()
             elif key == "STANDBY":
+                self._write_str(key, 3)
+                self.stdscr.refresh()
                 self._session.standby()
-                self._session.loop.stop()
+                self._session.loop.call_soon_threadsafe(self._session.loop.stop)
                 sys.exit()
+            self.controller.button(key, "press")
+        return key
