@@ -1,41 +1,69 @@
+"""Remote Play Session."""
 import asyncio
 import logging
 import socket
-import threading
 import time
 from base64 import b64decode, b64encode
-from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
+from concurrent.futures import ThreadPoolExecutor
 from enum import IntEnum
 from functools import partial
 from struct import pack_into
 
 import requests
 from Cryptodome.Random import get_random_bytes
+from pyee import ExecutorEventEmitter
 
 from .av import AVHandler
-from .const import (FPS, OS_TYPE, RP_CRYPT_SIZE, RP_PORT, RP_VERSION, TYPE_PS4,
-                    TYPE_PS5, USER_AGENT, Resolution)
+from .const import (
+    FPS,
+    OS_TYPE,
+    RP_CRYPT_SIZE,
+    RP_PORT,
+    RP_VERSION_PS4,
+    RP_VERSION_PS5,
+    TYPE_PS4,
+    TYPE_PS5,
+    USER_AGENT,
+    Resolution,
+    StreamType,
+)
 from .crypt import SessionCipher
-from .ddp import get_status, wakeup
+from .ddp import async_get_status, async_wakeup
 from .errors import RemotePlayError, RPErrorHandler
-from .feedback import Controller
-from .keys import SESSION_KEY_0, SESSION_KEY_1
+from .keys import (
+    SESSION_KEY_0_PS4,
+    SESSION_KEY_1_PS4,
+    SESSION_KEY_0_PS5,
+    SESSION_KEY_1_PS5,
+)
 from .stream import RPStream
-from .util import format_regist_key, listener, log_bytes
+from .util import format_regist_key, log_bytes
 
 _LOGGER = logging.getLogger(__name__)
 
-RP_INIT_URL = "/sie/ps4/rp/sess/init"
-RP_SESSION_URL = "/sie/ps4/rp/sess/ctrl"
+RP_INIT_URL = "/sie/{}/rp/sess/init"
+RP_SESSION_URL = "/sie/{}/rp/sess/ctrl"
 
-DID_PREFIX = b'\x00\x18\x00\x00\x00\x07\x00\x40\x00\x80'
+
+DID_PREFIX = b"\x00\x18\x00\x00\x00\x07\x00\x40\x00\x80"
 
 HEARTBEAT_RESPONSE = b"\x00\x00\x00\x00\x01\xfe\x00\x00"
 
 RP_ERROR = RPErrorHandler()
 
+HOST_TYPES = {
+    TYPE_PS4: {
+        "keys": (SESSION_KEY_0_PS4, SESSION_KEY_1_PS4),
+        "version": RP_VERSION_PS4,
+    },
+    TYPE_PS5: {
+        "keys": (SESSION_KEY_0_PS5, SESSION_KEY_1_PS5),
+        "version": RP_VERSION_PS5,
+    },
+}
 
-def _get_headers(host: str, regist_key: str) -> dict:
+
+def _get_headers(host_type: str, host: str, regist_key: str) -> dict:
     """Return headers."""
     headers = {
         "Host": f"{host}:{RP_PORT}",
@@ -43,12 +71,20 @@ def _get_headers(host: str, regist_key: str) -> dict:
         "Connection": "close",
         "Content-Length": "0",
         "RP-Registkey": regist_key,
-        "Rp-Version": RP_VERSION,
+        "Rp-Version": HOST_TYPES[host_type]["version"],
     }
     return headers
 
 
-def _get_session_headers(host: str, auth: str, did: str, os_type: str, bitrate: str):
+def _get_session_headers(
+    host_type: str,
+    host: str,
+    auth: str,
+    did: str,
+    os_type: str,
+    bitrate: str,
+    stream_type: str,
+):
     """Return Connect Headers."""
     headers = {
         "Host": f"{host}:{RP_PORT}",
@@ -56,7 +92,7 @@ def _get_session_headers(host: str, auth: str, did: str, os_type: str, bitrate:
         "Connection": "keep-alive",
         "Content-Length": "0",
         "RP-Auth": auth,
-        "RP-Version": RP_VERSION,
+        "RP-Version": HOST_TYPES[host_type]["version"],
         "RP-Did": did,
         "RP-ControllerType": "3",
         "RP-ClientType": "11",
@@ -64,16 +100,36 @@ def _get_session_headers(host: str, auth: str, did: str, os_type: str, bitrate:
         "RP-ConPath": "1",
         "RP-StartBitrate": bitrate,
     }
-    _LOGGER.debug("Session Headers: %s", headers)
+    if host_type == TYPE_PS5:
+        headers["RP-StreamingType"] = stream_type
     return headers
 
 
-def _get_rp_nonce(nonce: bytes) -> bytes:
+def _get_stream_type(stream_type: StreamType) -> bytes:
+    """Return Stream Type."""
+    stream_type = int(stream_type)
+    return bytes(
+        bytearray(
+            [
+                stream_type & 255,
+                (stream_type >> 8) & 255,
+                (stream_type >> 16) & 255,
+                (stream_type >> 24) & 255,
+            ]
+        )
+    )
+
+
+def _get_rp_nonce(host_type: str, nonce: bytes) -> bytes:
     """Return RP nonce."""
+    session_key = HOST_TYPES[host_type]["keys"][0]
     rp_nonce = bytearray(RP_CRYPT_SIZE)
-    key = SESSION_KEY_0[((nonce[0] >> 3) * 112):]
+    key = session_key[((nonce[0] >> 3) * 112) :]
     for index in range(0, RP_CRYPT_SIZE):
-        shift = nonce[index] + 54 + index
+        if host_type == TYPE_PS5:
+            shift = nonce[index] - 45 - index
+        else:
+            shift = nonce[index] + 54 + index
         shift ^= key[index]
         rp_nonce[index] = shift % 256
     rp_nonce = bytes(rp_nonce)
@@ -81,13 +137,19 @@ def _get_rp_nonce(nonce: bytes) -> bytes:
     return rp_nonce
 
 
-def _get_aes_key(nonce: bytes, rp_key: bytes) -> bytes:
+def _get_aes_key(host_type: str, nonce: bytes, rp_key: bytes) -> bytes:
     """Return AES key."""
     aes_key = bytearray(16)
-    key = SESSION_KEY_1[((nonce[7] >> 3) * 112):]
+    session_key = HOST_TYPES[host_type]["keys"][1]
+    key = session_key[((nonce[7] >> 3) * 112) :]
     for index in range(0, RP_CRYPT_SIZE):
-        shift = (key[index] ^ rp_key[index]) + 33 + index
-        shift ^= nonce[index]
+        if host_type == TYPE_PS5:
+            shift = rp_key[index] + 24 + index
+            shift ^= nonce[index]
+            shift ^= key[index]
+        else:
+            shift = (key[index] ^ rp_key[index]) + 33 + index
+            shift ^= nonce[index]
         aes_key[index] = shift % 256
     aes_key = bytes(aes_key)
     log_bytes("AES Key", aes_key)
@@ -101,8 +163,9 @@ def _gen_did() -> bytes:
     return did
 
 
-class Session():
-    """Controller for RP Session."""
+class Session:
+    """RP Session Async."""
+
     STATE_INIT = "init"
     STATE_READY = "ready"
     STATE_STOP = "stop"
@@ -110,12 +173,13 @@ class Session():
 
     class MessageType(IntEnum):
         """Enum for Message Types."""
+
         LOGIN_PIN_REQUEST = 0x04
         LOGIN_PIN_RESPONSE = 0x8004
         LOGIN = 0x05
         SESSION_ID = 0x33
-        HEARTBEAT_REQUEST = 0xfe
-        HEARTBEAT_RESPONSE = 0x1fe
+        HEARTBEAT_REQUEST = 0xFE
+        HEARTBEAT_RESPONSE = 0x1FE
         STANDBY = 0x50
         KEYBOARD_ENABLE_TOGGLE = 0x20
         KEYBOARD_OPEN = 0x21
@@ -124,6 +188,26 @@ class Session():
         KEYBOARD_TEXT_CHANGE_RES = 0x24
         KEYBOARD_CLOSE_REQ = 0x25
 
+    class Protocol(asyncio.Protocol):
+        """Protocol for session."""
+
+        def __init__(self, session):
+            self.transport = None
+            self.session = session
+
+        def connection_made(self, transport):
+            """Callback for connection made."""
+            _LOGGER.debug("Connected")
+            self.transport = transport
+
+        def data_received(self, data):
+            """Callback for data received."""
+            self.session.handle(data)
+
+        def close(self):
+            """Close Transport."""
+            self.transport.close()
+
     def __repr__(self):
         return (
             f"<RP Session host={self._host} "
@@ -131,11 +215,27 @@ class Session():
             f"fps={self.fps}>"
         )
 
-    def __init__(self, host: str, profile: dict, resolution="720p", fps="high", av_receiver=None, use_hw=False, quality="default", **kwargs):
+    def __del__(self):
+        self.stop()
+
+    def __init__(
+        self,
+        host: str,
+        profile: dict,
+        av_receiver=None,
+        loop=None,
+        resolution="360p",
+        fps="low",
+        quality="very_low",
+        codec="h264",
+        hdr=False,
+        **kwargs,
+    ):
         self._host = host
         self._profile = profile
         self._regist_data = {}
-        self._session_id = b''
+        self._session_id = b""
+        self._server_type = None
         self._type = ""
         self._mac_address = ""
         self._name = ""
@@ -147,16 +247,24 @@ class Session():
         self._cipher = None
         self._state = Session.STATE_INIT
         self._stream = None
+        self._worker = None
         self._kwargs = kwargs
+        self._hdr = hdr
+        self._codec = codec
         self.quality = quality
-        self.use_hw = use_hw
         self.max_width = self.max_height = None
         self.fps = FPS.preset(fps)
         self.resolution = Resolution.preset(resolution)
         self.error = ""
-        self.av_receiver = av_receiver(self) if av_receiver is not None else None
+        self.av_receiver = av_receiver
         self.av_handler = AVHandler(self)
-        self.controller = Controller(self)
+        self.events = ExecutorEventEmitter()
+
+        self.loop = asyncio.get_event_loop() if loop is None else loop
+        self._protocol = None
+        self._transport = None
+        self._tasks = []
+        self._thread_executor = ThreadPoolExecutor()
 
         self._ready_event = None
         self._stop_event = None
@@ -169,7 +277,7 @@ class Session():
             return False
         self._regist_key = self._kwargs.get("regist_key")
         try:
-            self._rp_key = bytes.fromhex(self._keargs.get("rp_key"))
+            self._rp_key = bytes.fromhex(self._kwargs.get("rp_key"))
         except ValueError:
             _LOGGER.error("Invalid RP Key")
             return False
@@ -193,17 +301,13 @@ class Session():
         self._rp_key = bytes.fromhex(self._regist_data["RP-Key"])
         return True
 
-    def _get_status(self) -> dict:
-        """Return dict of device status."""
-        return get_status(self._host, host_type=self.type)
-
-    def _check_host(self) -> tuple:
+    async def _check_host(self) -> tuple:
         """Return True, True if host is available."""
-        device = self._get_status()
+        device = await async_get_status(self._host, host_type=self.type)
         if not device:
             _LOGGER.error("Could not detect host at: %s", self._host)
             return (False, False, None)
-        if device.get("status_code") != 200:
+        if device.get("status-code") != 200:
             _LOGGER.info("Host: %s is not on", self._host)
             return (True, False, device)
         return (True, True, device)
@@ -211,15 +315,19 @@ class Session():
     def _get_rp_url(self, request_type: str) -> str:
         valid_types = ["init", "session"]
         if request_type not in valid_types:
-            raise ValueError("Unknown ")
+            raise RemotePlayError("Unknown request type")
         url_slug = RP_INIT_URL if request_type == "init" else RP_SESSION_URL
+        url_slug = url_slug.format(self.type.lower())
         url = f"http://{self.host}:{RP_PORT}{url_slug}"
         return url
 
-    def _send_auth_request(self, request_type: str, headers: dict, stream: bool) -> requests.models.Response:
+    def _send_auth_request(
+        self, request_type: str, headers: dict, stream: bool
+    ) -> requests.models.Response:
         """Return response. Send Auth Request."""
         response = None
         url = self._get_rp_url(request_type)
+        # Using requests as it is easier to get the socket later.
         response = requests.get(url, headers=headers, stream=stream, timeout=3)
         if response is None:
             _LOGGER.error("Timeout: Auth")
@@ -228,7 +336,6 @@ class Session():
     def _parse_init(self, response: requests.models.Response) -> bytes:
         """Return nonce. Parse init response."""
         nonce = None
-        _LOGGER.debug(response.headers)
         if response.status_code != 200:
             reason = response.headers.get("RP-Application-Reason")
             reason = int.from_bytes(bytes.fromhex(reason), "big")
@@ -246,16 +353,22 @@ class Session():
 
     def _get_session_headers(self, nonce: bytes) -> dict:
         """Return Session headers."""
-        rp_nonce = _get_rp_nonce(nonce)
-        aes_key = _get_aes_key(nonce, self._rp_key)
-        self._cipher = SessionCipher(aes_key, rp_nonce, counter=0)
+        stream_type = _get_stream_type(self.stream_type)
+        rp_nonce = _get_rp_nonce(self.type, nonce)
+        aes_key = _get_aes_key(self.type, nonce, self._rp_key)
+        self._cipher = SessionCipher(self.type, aes_key, rp_nonce, counter=0)
 
-        regist_key = b''.join([bytes.fromhex(self._regist_key), bytes(8)])
+        regist_key = b"".join([bytes.fromhex(self._regist_key), bytes(8)])
         auth = b64encode(self._cipher.encrypt(regist_key)).decode()
         did = b64encode(self._cipher.encrypt(_gen_did())).decode()
-        os_type = b64encode(self._cipher.encrypt(OS_TYPE.encode().ljust(10, b'\x00'))).decode()
+        os_type = b64encode(
+            self._cipher.encrypt(OS_TYPE.encode().ljust(10, b"\x00"))
+        ).decode()
         bitrate = b64encode(self._cipher.encrypt(bytes(4))).decode()
-        return _get_session_headers(self.host, auth, did, os_type, bitrate)
+        stream_type = b64encode(self._cipher.encrypt(stream_type)).decode()
+        return _get_session_headers(
+            self.type, self.host, auth, did, os_type, bitrate, stream_type
+        )
 
     def _authenticate(self, nonce: bytes) -> bool:
         """Return True if successful. Send Session Auth."""
@@ -267,14 +380,18 @@ class Session():
         server_type = response.headers.get("RP-Server-Type")
         if response.status_code != 200 or server_type is None:
             return False
-        self._server_type = int.from_bytes(self._cipher.decrypt(b64decode(server_type)), 'little')
+        self._server_type = int.from_bytes(
+            self._cipher.decrypt(b64decode(server_type)), "little"
+        )
         _LOGGER.debug("Server Type: %s", self._server_type)
         self._sock = socket.fromfd(
-            response.raw.fileno(), socket.AF_INET, socket.SOCK_STREAM)
+            response.raw.fileno(), socket.AF_INET, socket.SOCK_STREAM
+        )
         return True
 
-    def _handle(self, data: bytes):
+    def handle(self, data: bytes):
         """Handle Data."""
+
         def invalid_session_id(session_id: bytes) -> bytes:
             """Return a valid session id.
 
@@ -287,7 +404,7 @@ class Session():
             session_id = bytearray(session_id)
             for char in session_id:
                 new_id.append(chr(char).encode())
-            new_id = b''.join(new_id)
+            new_id = b"".join(new_id)
             log_bytes("New Session ID", new_id)
             return new_id
 
@@ -323,10 +440,10 @@ class Session():
             self._ready_event.set()
 
         if time.time() - self._hb_last > 5:
-            _LOGGER.info("Session HB Timeout. Sending HB")
+            _LOGGER.debug("Session HB Timeout. Sending HB")
             self._send_hb_request()
 
-    def _build_msg(self, msg_type: int, payload=b'') -> bytes:
+    def _build_msg(self, msg_type: int, payload=b"") -> bytes:
         """Return Message."""
         payload_size = len(payload)
         self._cipher.encrypt(payload)
@@ -335,15 +452,24 @@ class Session():
         return bytes(buf)
 
     def _send_hb_response(self):
-        msg = self._build_msg(Session.MessageType.HEARTBEAT_RESPONSE, HEARTBEAT_RESPONSE)
+        msg = self._build_msg(
+            Session.MessageType.HEARTBEAT_RESPONSE, HEARTBEAT_RESPONSE
+        )
         self.send(msg)
 
     def _send_hb_request(self):
         msg = self._build_msg(Session.MessageType.HEARTBEAT_REQUEST)
         self.send(msg)
 
+    def encrypt(self, data: bytes, counter: int = None):
+        """Return Encypted Data."""
+        if not self._cipher:
+            raise RemotePlayError("Session cipher not created")
+        return self._cipher.encrypt(data, counter=counter)
+
     def standby(self):
         """Set host to standby."""
+
         msg = self._build_msg(Session.MessageType.STANDBY)
         self.send(msg)
         _LOGGER.info("Sending Standby")
@@ -351,53 +477,56 @@ class Session():
 
     def send(self, data: bytes):
         """Send Data."""
-        self._sock.send(data)
+        self._protocol.transport.write(data)
         log_bytes("Session Send", data)
 
-    def wakeup(self):
+    async def wakeup(self):
         """Wakeup Host."""
         regist_key = format_regist_key(self._regist_key)
-        wakeup(self.host, regist_key, host_type=self.type)
+        await async_wakeup(self.host, regist_key, host_type=self.type)
 
-    def start(self, wakeup=True, autostart=True) -> bool:
+    async def start(self, wakeup=True, autostart=True) -> bool:
         """Start Session/RP Session."""
-        self._ready_event = threading.Event()
-        self._stop_event = threading.Event()
-        self.receiver_started = threading.Event()
-        self.stream_ready = threading.Event()
-        status = self._check_host()
+        _LOGGER.info("Session Started")
+        self._ready_event = asyncio.Event()
+        self._stop_event = asyncio.Event()
+        self.receiver_started = asyncio.Event()
+        self.stream_ready = asyncio.Event()
+
+        _LOGGER.debug("Running Async")
+        status = await self._check_host()
         if not status[0]:
             self.error = f"Host @ {self._host} is not reachable."
             return False
         if not self._init_profile(status[2]):
             self.error = "Profile is not registered with host"
-            if not self._init_profile_kwargs(status[2]):
-                return False
+            return False
         if not status[1]:
             if wakeup:
-                self.wakeup()
+                await self.wakeup()
                 self.error = "Host is in Standby. Attempting to wakeup."
             return False
-        if not self.connect():
+        if not await self.run_io(self.connect):
             _LOGGER.error("Session Auth Failed")
+            if not self.error:
+                self.error = "Auth Failed."
             return False
         _LOGGER.info("Session Auth Success")
         self._state = Session.STATE_READY
-        self._worker = threading.Thread(
-            target=listener,
-            args=("Session", self._sock, self._handle, self._stop_event),
+        _, self._protocol = await self.loop.connect_accepted_socket(
+            lambda: Session.Protocol(self), self._sock
         )
-        self._worker.start()
-        self._ready_event.wait()
+        await self._ready_event.wait()
         if autostart:
             self.start_stream()
         return True
 
     def connect(self) -> bool:
         """Connect to Host."""
-        headers = _get_headers(self.host, self._regist_key)
+        headers = _get_headers(self.type, self.host, self._regist_key)
         response = self._send_auth_request("init", headers, stream=False)
         if response is None:
+            _LOGGER.error("No response for Init")
             return False
         nonce = self._parse_init(response)
         if nonce is None:
@@ -417,26 +546,71 @@ class Session():
         if not self.session_id:
             _LOGGER.error("Session ID not received")
             return
-        stop_event = self._stop_event if not test else threading.Event()
+        stop_event = self._stop_event if not test else asyncio.Event()
         cb_stop = self._cb_stop_test if test else None
         if not test and self.av_receiver:
             self.av_handler.add_receiver(self.av_receiver)
             _LOGGER.info("Waiting for Receiver...")
-            self.receiver_started.wait()
-        self._stream = RPStream(self, stop_event, is_test=test, cb_stop=cb_stop, mtu=mtu, rtt=rtt)
-        self._stream.connect()
+            self.loop.create_task(self.receiver_started.wait())
+        self._stream = RPStream(
+            self, stop_event, is_test=test, cb_stop=cb_stop, mtu=mtu, rtt=rtt
+        )
+        self.loop.create_task(self._stream.async_connect())
+        if test:
+            self.loop.create_task(self.wait_for_test(stop_event))
+
+    async def wait_for_test(self, stop_event):
+        """Wait for network test to complete. Uses defaults if timed out."""
+        try:
+            await asyncio.wait_for(stop_event.wait(), timeout=3.0)
+        except asyncio.exceptions.TimeoutError:
+            _LOGGER.warning("Network Test timed out. Using Default MTU and RTT")
+            self._stream.stop()
+
+    def init_av_handler(self):
+        """Run AV Handler."""
+        self._tasks.append(self.loop.create_task(self.run_io(self.av_handler.worker)))
 
     def stop(self):
-        """Stop Stream."""
+        """Stop Session."""
         if self.state == Session.STATE_STOP:
             _LOGGER.debug("Session already stopping")
             return
-        _LOGGER.info("Session Received Stop Signal")
-        if self._stop_event:
-            self._stop_event.set()
+        _LOGGER.debug("Session Received Stop Signal")
+        if self._stream:
+            self._stream.stop()
+        self._stop_event.set()
+        if self._tasks:
+            for task in self._tasks:
+                task.cancel()
+        if self._thread_executor:
+            self._thread_executor.shutdown()
+        if self._protocol:
+            self._protocol.close()
+        if self.events:
+            self.events.emit("stop")
+            self.events.remove_all_listeners()
 
-    def init_controller(self):
-        self.controller.start()
+        self._tasks = []
+        self._stream = None
+        self._thread_executor = None
+        self._protocol = None
+        self.av_handler = None
+        self.events = None
+
+    async def run_io(self, func, *args, **kwargs):
+        """Run blocking function in executor."""
+        if not self._thread_executor:
+            if not self.is_stopped:
+                _LOGGER.warning("No Executor and session is not stopped")
+            return None
+        return await self.loop.run_in_executor(
+            self._thread_executor, partial(func, *args, **kwargs)
+        )
+
+    def sync_run_io(self, func, *args, **kwargs):
+        """Run blocking function in executor. Called from sync method."""
+        asyncio.ensure_future(self.run_io(func, *args, **kwargs))
 
     @property
     def host(self) -> str:
@@ -470,7 +644,7 @@ class Session():
     @property
     def is_running(self) -> bool:
         """Return True if running."""
-        return self.state != Session.STATE_STOP
+        return self.state == Session.STATE_READY
 
     @property
     def is_stopped(self) -> bool:
@@ -483,123 +657,34 @@ class Session():
         return self._session_id
 
     @property
-    def video_format(self):
-        formats = {
-            TYPE_PS4: "h264",
-            TYPE_PS5: "h265",
-        }
-        return formats.get(self.type)
-    
+    def stream(self):
+        """Return Stream."""
+        return self._stream
 
-class SessionAsync(Session):
+    @property
+    def stop_event(self):
+        """Return Stop Event."""
+        return self._stop_event
 
-    class Protocol(asyncio.Protocol):
+    @property
+    def video_format(self) -> str:
+        """Return video format"""
+        return self._codec.lower()
 
-        def __init__(self, session):
-            self.transport = None
-            self.session = session
+    @property
+    def hdr(self) -> bool:
+        """Return True if HDR."""
+        return self._hdr and "h264" not in self._codec
 
-        def connection_made(self, transport):
-            _LOGGER.debug("Connected")
-            self.transport = transport
-
-        def data_received(self, data):
-            self.session._handle(data)
-
-        def close(self):
-            self.transport.close()
-
-    def __init__(self, host: str, profile: dict, resolution="720p", fps="high", av_receiver=None, use_hw=False, quality="default", loop=None, **kwargs):
-        super().__init__(host, profile, resolution, fps, av_receiver, use_hw, quality, **kwargs)
-        self.loop = asyncio.get_event_loop() if loop is None else loop
-        self._protocol = None
-        self._transport = None
-        self._tasks = []
-        self._thread_executor = ThreadPoolExecutor()
-
-    async def run_io(self, func, *args, **kwargs):
-        return await self.loop.run_in_executor(self._thread_executor, partial(func, *args, **kwargs))
-
-    def sync_run_io(self, func, *args, **kwargs):
-        asyncio.ensure_future(self.run_io(func, *args, **kwargs))
-
-    async def start(self, wakeup=True, autostart=True) -> bool:
-        """Start Session/RP Session."""
-        _LOGGER.info("Session Started")
-        self._ready_event = asyncio.Event()
-        self._stop_event = asyncio.Event()
-        self.receiver_started = asyncio.Event()
-        self.stream_ready = asyncio.Event()
-
-        _LOGGER.debug("Running Async")
-        status = await self.run_io(self._check_host)
-        if not status[0]:
-            self.error = f"Host @ {self._host} is not reachable."
-            return False
-        if not self._init_profile(status[2]):
-            self.error = "Profile is not registered with host"
-            return False
-        if not status[1]:
-            if wakeup:
-                await self.run_io(self.wakeup)
-                self.error = "Host is in Standby. Attempting to wakeup."
-            return False
-        if not await self.run_io(self.connect):
-            _LOGGER.error("Session Auth Failed")
-            return False
-        _LOGGER.info("Session Auth Success")
-        self._state = Session.STATE_READY
-        _, self._protocol = await self.loop.connect_accepted_socket(lambda: SessionAsync.Protocol(self), self._sock)
-        await self._ready_event.wait()
-        if autostart:
-            self.start_stream()
-        return True
-
-    def send(self, data: bytes):
-        """Send Data."""
-        self._protocol.transport.write(data)
-        log_bytes(f"Session Send", data)
-
-    def start_stream(self, test=True, mtu=None, rtt=None):
-        """Start Stream."""
-        if not self.session_id:
-            _LOGGER.error("Session ID not received")
-            return
-        stop_event = self._stop_event if not test else asyncio.Event()
-        cb_stop = self._cb_stop_test if test else None
-        if not test and self.av_receiver:
-            self.av_handler.add_receiver(self.av_receiver)
-            _LOGGER.info("Waiting for Receiver...")
-            self.loop.create_task(self.receiver_started.wait())
-        self._stream = RPStream(self, stop_event, is_test=test, cb_stop=cb_stop, mtu=mtu, rtt=rtt)
-        self.loop.create_task(self._stream.async_connect())
-        if test:
-            self.loop.create_task(self.wait_for_test(stop_event))
-        else:
-            self._tasks.append(self.loop.create_task(self.run_io(self.controller.worker)))
-
-    async def wait_for_test(self, stop_event):
+    @property
+    def stream_type(self) -> StreamType:
+        """Return Stream Type."""
+        _type = self.video_format.upper()
+        _type = _type.split("_")[0]  # Remove codec suffix
+        if self.hdr:
+            _type = f"{_type}_HDR"
         try:
-            await asyncio.wait_for(stop_event.wait(), timeout=3.0)
-        except asyncio.exceptions.TimeoutError:
-            _LOGGER.warning("Network Test timed out. Using Default MTU and RTT")
-            self._stream.stop()
-
-    def init_av_handler(self):
-        self._tasks.append(self.loop.create_task(self.run_io(self.av_handler.worker)))
-
-    def stop(self):
-        """Stop Stream."""
-        if self.state == Session.STATE_STOP:
-            _LOGGER.debug("Session already stopping")
-            return
-        _LOGGER.info("Session Received Stop Signal")
-        if self._stream:
-            self._stream.stop()
-        self._stop_event.set()
-        if self._tasks:
-            for task in self._tasks:
-                task.cancel()
-        self._thread_executor.shutdown()
-        if self._protocol:
-            self._protocol.close()
+            stream_type = StreamType[_type]
+        except KeyError:
+            stream_type = StreamType["H264"]
+        return stream_type
