@@ -7,14 +7,11 @@ import select
 import time
 from binascii import hexlify
 
-from pyee import ExecutorEventEmitter
 
 from .const import CONTROLS_FILE, OPTIONS_FILE, PROFILE_DIR, PROFILE_FILE
 
 _LOGGER = logging.getLogger(__name__)
 
-event_emitter = ExecutorEventEmitter()
-
 
 def check_dir() -> pathlib.Path:
     """Return path. Check file dir and create dir if not exists."""
@@ -27,7 +24,7 @@ def check_dir() -> pathlib.Path:
 def check_file(path: pathlib.Path):
     """Check if file exists and create."""
     if not path.is_file():
-        with open(path, "w") as _file:
+        with open(path, "w", encoding="utf-8") as _file:
             json.dump({}, _file)
 
 
@@ -40,7 +37,7 @@ def get_mapping(path=None) -> dict:
     else:
         path = pathlib.Path(path)
     check_file(path)
-    with open(path, "r") as _file:
+    with open(path, "r", encoding="utf-8") as _file:
         data = json.load(_file)
     return data
 
@@ -51,7 +48,7 @@ def write_mapping(mapping: dict, path=None):
         path = pathlib.Path.home() / PROFILE_DIR / CONTROLS_FILE
     else:
         path = pathlib.Path(path)
-    with open(path, "w") as _file:
+    with open(path, "w", encoding="utf-8") as _file:
         json.dump(mapping, _file)
 
 
@@ -64,7 +61,7 @@ def get_options(path=None) -> dict:
     else:
         path = pathlib.Path(path)
     check_file(path)
-    with open(path, "r") as _file:
+    with open(path, "r", encoding="utf-8") as _file:
         data = json.load(_file)
     return data
 
@@ -75,7 +72,7 @@ def write_options(options: dict, path=None):
         path = pathlib.Path.home() / PROFILE_DIR / OPTIONS_FILE
     else:
         path = pathlib.Path(path)
-    with open(path, "w") as _file:
+    with open(path, "w", encoding="utf-8") as _file:
         json.dump(options, _file)
 
 
@@ -88,7 +85,7 @@ def get_profiles(path=None) -> list:
     else:
         path = pathlib.Path(path)
     check_file(path)
-    with open(path, "r") as _file:
+    with open(path, "r", encoding="utf-8") as _file:
         data = json.load(_file)
     return data
 
@@ -99,7 +96,7 @@ def write_profiles(profiles: dict, path=None):
         path = pathlib.Path.home() / PROFILE_DIR / PROFILE_FILE
     else:
         path = pathlib.Path(path)
-    with open(path, "w") as _file:
+    with open(path, "w", encoding="utf-8") as _file:
         json.dump(profiles, _file)
 
 
@@ -120,30 +117,59 @@ def add_profile(profiles: dict, user_data: dict) -> dict:
     return profiles
 
 
+def get_users(device_id, profiles=None, path=None):
+    """Return users for device."""
+    users = []
+    if not profiles:
+        profiles = get_profiles(path)
+    for user, data in profiles.items():
+        hosts = data.get("hosts")
+        if not hosts:
+            continue
+        if hosts.get(device_id):
+            users.append(user)
+    return users
+
+
 def add_regist_data(profile: dict, host: dict, data: dict) -> dict:
     """Add regist data to profile and return profile."""
-    mac_address = host['host-id']
-    host_type = host['host-type']
+    mac_address = host["host-id"]
+    host_type = host["host-type"]
     for key in list(data.keys()):
         if key.startswith(host_type):
             value = data.pop(key)
             new_key = key.split("-")[1]
             data[new_key] = value
-    profile['hosts'][mac_address] = {'data': data, 'type': host_type}
+    profile["hosts"][mac_address] = {"data": data, "type": host_type}
     return profile
 
 
 def format_regist_key(regist_key: str) -> bytes:
     """Format Regist Key for wakeup."""
-    regist_key = int.from_bytes(bytes.fromhex(bytes.fromhex(regist_key).decode()), "big")
+    regist_key = int.from_bytes(
+        bytes.fromhex(bytes.fromhex(regist_key).decode()), "big"
+    )
     return regist_key
 
 
+def get_devices(path=None) -> dict:
+    """Return dict of devices from profiles."""
+    devices = {}
+    profiles = get_profiles(path)
+    for _, data in profiles.items():
+        _device_data = data.get("hosts")
+        if not _device_data:
+            continue
+        devices.update(_device_data)
+    return devices
+
+
 def log_bytes(name: str, data: bytes):
     """Log bytes."""
     mod = inspect.getmodulename(inspect.stack()[1].filename)
     logging.getLogger(f"{__package__}.{mod}").debug(
-        "Length: %s, %s: %s", len(data), name, hexlify(data))
+        "Length: %s, %s: %s", len(data), name, hexlify(data)
+    )
 
 
 def from_b(_bytes: bytes, order="big") -> int:
@@ -172,16 +198,20 @@ def listener(name: str, sock, handle, stop_event):
         time.sleep(0.001)
 
     sock.close()
-    _LOGGER.info(f"{name} Stopped")
+    _LOGGER.info("%s Stopped", name)
 
 
 def timeit(func):
     """Time Function."""
+
     def inner(*args, **kwargs):
         start = time.time()
         result = func(*args, **kwargs)
         end = time.time()
         elapsed = round(end - start, 8)
-        _LOGGER.info("Timed %s.%s at %s seconds", func.__module__, func.__name__, elapsed)
+        _LOGGER.info(
+            "Timed %s.%s at %s seconds", func.__module__, func.__name__, elapsed
+        )
         return result
+
     return inner
